# Natural Language Toolkit: Unit Tests
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Edward Loper <edloper@gradient.cis.upenn.edu>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
Unit tests for the NLTK modules.  These tests are intended to ensure
that changes that we make to NLTK's code don't accidentally introduce
bugs.

Use doctest_driver.py to run the tests::

  doctest_driver.py --help

NB. Popular options for NLTK documentation are::

  --ellipsis --normalize_whitespace

"""
